import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Hashtable;
import com.ibm.mq.*;

/**
 * Program Name
 *  MQPortScan
 *
 * Description
 *  This java class will attempt connect to a remote queue manager
 *  using various channels on a range of port numbers.
 *
 * Sample Command Line Parameters
 *  -h 127.0.0.1 -s Starting_Port_Number -e Ending_Port_Number
 *
 *  License: Apache Software License, Version 2
 *
 * @version 1.0.0
 * @author Roger Lacroix, Capitalware Inc.
 */
public class MQPortScan
{
   private String    userHostName   = null;
   private int       startPortNum   = 1414;
   private int       endPortNum     = 10000;
   private SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

   /*
    * Add user defined channels after system channels.
    */
   private final static String[]  ALL_CHANNELS_TO_BE_TESTED =
   {
      "SYSTEM.DEF.SVRCONN",
      "SYSTEM.AUTO.SVRCONN",
      "SYSTEM.ADMIN.SVRCONN",
      // User defined Channels
      "ADMIN.SVRCONN",
      "SSL.SVRCONN",
      "TEST.SVRCONN",
      "TST.SVRCONN",
      "DEV.SVRCONN",
      "QA.SVRCONN",
      "PROD.SVRCONN",
      "SVRCONN.CHL",
      "ADMIN.SVRCONN.CHL",
      "SSL.SVRCONN.CHL",
      "TEST.SVRCONN.CHL",
      "TST.SVRCONN.CHL",
      "DEV.SVRCONN.CHL",
      "QA.SVRCONN.CHL",
      "PROD.SVRCONN.CHL"
   };
   /* Set length as a constant to speed up loops */
   private final static int      CHL_COUNT = ALL_CHANNELS_TO_BE_TESTED.length;

   private final static String   OUTPUT_FILENAME        = "MQPortScan.csv";
   private final static String   OUTPUT_FILENAME_BACKUP = "MQPortScan.csv.bk";
   private final static String   CSV_HEADER = "QMgrName,Version,ChannelName,hostname/IP_Address,Port";

   /*
    * Set to true to wait on port scan where a server has a firewall.
    * This will add a significant amount of time to the run!!!!
    */
   private final static boolean waitFirewall = false;

   /*
    * The constructor
    */
   public MQPortScan()
   {
      super();
   }

   /**
    * Extract the command-line parameters and initialize the MQ variables.
    * @param args
    * @throws IllegalArgumentException
    */
   private void init(String[] args) throws IllegalArgumentException
   {
      Hashtable params = new Hashtable();
      if (args.length > 0 && (args.length % 2) == 0)
      {
         for (int i = 0; i < args.length; i += 2)
         {
            params.put(args[i], args[i + 1]);
         }
      }
      else
      {
         throw new IllegalArgumentException("Error: Mismatch keywords and values.");
      }

      if ( params.containsKey("-h") )
      {
         userHostName = (String) params.get("-h");

         if (userHostName.indexOf('*') != -1)
         {
            String[] ip = userHostName.replace('.', ':').split(":");
            if (ip[0].indexOf('*') != -1)
            {
               throw new IllegalArgumentException("Error: MQPortScan does not support '*' in the first part of the IP address.");
            }
         }
      }
      else
      {
         throw new IllegalArgumentException("Error: Missing '-h' keyword and value.");
      }

      try
      {
         if (params.get("-s") != null)
            startPortNum = Integer.parseInt((String) params.get("-s"));
      }
      catch (NumberFormatException e)
      {
         throw new IllegalArgumentException("Error: Invalid number for '-s' keyword.");
      }

      try
      {
         if (params.get("-e") != null)
            endPortNum = Integer.parseInt((String) params.get("-e"));
      }
      catch (NumberFormatException e)
      {
         throw new IllegalArgumentException("Error: Invalid number for '-e' keyword.");
      }

      if ( (startPortNum > endPortNum) ||
           (startPortNum <= 0) ||
           (endPortNum <= 0) )
      {
         throw new IllegalArgumentException("Error: Starting and Ending port number must be greater than zero and Starting Port # must be less than Ending Port #");
      }

      // I don't want to see MQ exceptions at the console.
      MQException.log = null;
   }

   /**
    * Clean up CSV file and determine if the hostname has any wildcards.
    */
   private void processIt()
   {
      /*
       * Delete backup file if it exists.
       */
      try
      {
         if ( (new File(OUTPUT_FILENAME_BACKUP)).exists() )
         {
            if ( !((new File(OUTPUT_FILENAME_BACKUP)).delete()) )
               System.err.println("Error: The deletion of '" + OUTPUT_FILENAME_BACKUP + "' failed.");
         }
      }
      catch (Exception e)
      {
         System.err.println("Error: " +e.getLocalizedMessage());
      }

      /*
       * Rename CSV file.
       */
      try
      {
         if ((new File(OUTPUT_FILENAME)).exists())
         {
            File csvFile = new File(OUTPUT_FILENAME);
            if ( !(csvFile.renameTo(new File(OUTPUT_FILENAME_BACKUP))) )
               System.err.println("Error: The renaming of '" + OUTPUT_FILENAME + "' failed.");
         }
      }
      catch (Exception e)
      {
         System.err.println("Error: " +e.getLocalizedMessage());
      }

      /*
       * Write CSV header line.
       */
      writeFile(OUTPUT_FILENAME, CSV_HEADER.getBytes(), false);

      /*
       * Does the supplied hostname have any asterisks "*" ?
       */
      if (userHostName.indexOf('*') == -1)
         hitPorts(userHostName);
      else
         wildCardSearch(userHostName.replace('.', ':').split(":"));
   }

   /**
    * Search by wild card hostname.
    *
    * @param ip
    */
   private void wildCardSearch(String[] ip)
   {
      String[] newIP = (String[])ip.clone();

      if (ip[1].indexOf('*') != -1)
      {
         for (int x=1; x < 256; x++)
         {
            newIP[1] = x+"";
            wildCardSearch(newIP);
         }
      }
      else if (ip[2].indexOf('*') != -1)
      {
         for (int y=1; y < 256; y++)
         {
            newIP[2] = y+"";
            wildCardSearch(newIP);
         }
      }
      else if (ip[3].indexOf('*') != -1)
      {
         for (int z=1; z < 256; z++)
         {
            newIP[3] = z+"";
            hitPorts(newIP[0]+"."+newIP[1]+"."+newIP[2]+"."+newIP[3]);
         }
      }
   }

   /**
    * Loop through the ports of a given hostname or IP Address.
    *
    * @param hostname
    */
   private void hitPorts(String hostname)
   {
      boolean  hostAvailable = true;
      long     startTime;
      String   qMgrName = null;
      String   response = null;

      for (int i=startPortNum; (i <= endPortNum) && (hostAvailable || waitFirewall); i++)
      {
         System.out.println(dateFormat.format(new java.util.Date()) + " : " + "Scanning port # "+i + " of hostname " + hostname);
         for (int j=0; j < CHL_COUNT; j++)
         {
            startTime = timeIs();
            response = testConn(ALL_CHANNELS_TO_BE_TESTED[j], hostname, i);

            if ( (timeIs() - startTime) > 10000)
            {
               /*
                * Any thing over 10 seconds means there is no server/host at that IP address
                * or there is a firewall in the way.
                */
               if (!waitFirewall)
                  System.out.println(dateFormat.format(new java.util.Date()) + " : No server/host at " + hostname + " and Port # " + i);
               hostAvailable = false;
               break;
            }
            else if ("?".equals(response))
            {
               // Found something.
            }
            else if (response != null)
            {
               qMgrName = response;
            }
            else
            {
               // No queue manager.
               break;
            }
         } // for (int j=0;

         /*
          * A lot of MQAdmins create a channel in the following format, so lets try it:
          *   {QMgrName}.SVRCONN
          */
         if (qMgrName!= null)
            testConn(qMgrName+".SVRCONN", hostname, i);

         qMgrName = null;
      } // for (int i=startPortNum;
   }

   /**
    * Attempt to connect to a queue manager. If successful, output the info and disconnect.
    *
    * @param chlName Channel Name
    * @param hostname hostname or IP address
    * @param portNum port number
    */
   private String testConn(String chlName, String hostname, int portNum)
   {
      try
      {
         MQEnvironment.channel = chlName;
         MQEnvironment.hostname = hostname;
         MQEnvironment.port = portNum;

         MQQueueManager _qMgr = new MQQueueManager("");

         String qMgrName = _qMgr.getAttributeString(MQC.MQCA_Q_MGR_NAME, 48).trim();
         String outputLine = qMgrName + "," + _qMgr.getCommandLevel() + "," + MQEnvironment.channel + "," + MQEnvironment.hostname + "," + MQEnvironment.port;
         writeFile(OUTPUT_FILENAME, outputLine.getBytes(), true);

         try
         {
            _qMgr.disconnect();
         }
         catch (MQException e)
         {}

         return qMgrName;
      }
      catch (MQException e)
      {
         // System.out.println(dateFormat.format(new java.util.Date()) + " : " + "Completion Code = " + e.completionCode + " : Reason Code = " + e.reasonCode);
         if (e.reasonCode == MQException.MQRC_Q_MGR_NOT_AVAILABLE)
            return null;
         else if ( (e.reasonCode == MQException.MQRC_CONNECTION_BROKEN) ||
                   (e.reasonCode == MQException.MQRC_NOT_AUTHORIZED)    ||
                   (e.reasonCode == MQException.MQRC_SECURITY_ERROR)    )
            return "?";
         else
            return null;
      }
   }

   /**
    * This method will write data to a file.
    *
    * @param fileName the name of the file.
    * @param data the data to be written.
    * @param append flag
    */
   public void writeFile(String fileName, byte[] data, boolean append)
   {
      RandomAccessFile out;
      try
      {
         out = new RandomAccessFile(fileName, "rw");
         if (append)
         {
            long fileLen = out.length();
            out.seek(fileLen);   // move to the end of the file.
         }
         out.write(data);
         out.write("\n".getBytes());
         out.close();
      }
      catch (FileNotFoundException e)
      {
         System.err.println("Error: " +e.getLocalizedMessage());
      }
      catch (IOException e)
      {
         System.err.println("Error: " +e.getLocalizedMessage());
      }

      return;
   }

   /**
    * Get the current time.
    * @return long (timestamp)
    */
   private long timeIs()
   {
      Date currentDate = new java.util.Date();
      return currentDate.getTime();
   }

   /**
    * main line
    * @param args
    */
   public static void main(String[] args)
   {
      MQPortScan write = new MQPortScan();

      try
      {
         write.init(args);
         write.processIt();
      }
      catch (IllegalArgumentException e)
      {
         System.err.println("IllegalArgumentException: " + e.getLocalizedMessage());
         System.out.println("Usage: java MQPortScan -h host [-s Starting_PortNumber] [-e Ending_Port_Number]");
         System.exit(1);
      }

      System.exit(0);
   }
}
